#include <stdio.h>
#include <stdlib.h>
#include <netdb.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/bio.h>
#include <fcntl.h>

int check_for_errors (char *who)
{
  int err = ERR_get_error ();
  int err_lib = ERR_GET_LIB (err);
  int err_func = ERR_GET_FUNC (err);
  int err_reason = ERR_GET_REASON (err);

  if (err == SSL_ERROR_NONE)
    return err;
  fprintf (stderr, "%s: %s\n", who, ERR_error_string (err, NULL));
  fprintf (stderr, "  %s\n", ERR_lib_error_string (err_lib));
  fprintf (stderr, "  %s\n", ERR_func_error_string (err_func));
  fprintf (stderr, "  %s\n", ERR_reason_error_string (err_reason));
  return err;
}


int tcp_connect (char *host, int port)
{
  struct hostent *hp;
  struct sockaddr_in addr;
  int sock;
  /* int ofcmode; */

  if (! (hp=gethostbyname (host)))
    {
      printf ("couldn't resolve host\n");
      exit (1);
    }
  memset (&addr, 0, sizeof (addr));
  addr.sin_addr = * (struct in_addr*)
    hp->h_addr_list[ 0 ];
  addr.sin_family = AF_INET;
  addr.sin_port=htons (port);

  if ((sock=socket (AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
    {
      printf ("couldn't create socket\n");
      exit (1);
    }

  /*
  ofcmode=fcntl(sock,F_GETFL,0);
  ofcmode &= (~O_NDELAY);
  if (fcntl (sock, F_SETFL, ofcmode))
    {
      fprintf (stderr, "fcntl failed\n");
      exit (1);
    }
  */

  if (connect (sock, (struct sockaddr *)&addr, sizeof (addr)) < 0)
    {
      printf ("couldn't connect socket\n");
      exit (1);
    }

  return sock;
}


char *sd_state_to_string (int sd)
{
  if (sd == (SSL_SENT_SHUTDOWN | SSL_RECEIVED_SHUTDOWN))
    return "SSL_SENT_SHUTDOWN | SSL_RECEIVED_SHUTDOWN";
  else if (sd == SSL_SENT_SHUTDOWN)
    return "SSL_SENT_SHUTDOWN";
  else if (sd == SSL_RECEIVED_SHUTDOWN)
    return "SSL_RECEIVED_SHUTDOWN";
  else
    return "none";
}


int SSL_write_all (SSL *ssl, char *data, int len)
{
  int r;
  r = SSL_write (ssl, data, len);
  check_for_errors ("SSL_write");
  if (r < 0)
    {
      printf ("error from SSL_write\n");
      exit (1);
    }
  if (r < 0)
    {
      printf ("SSL_write returned 0\n");
      exit (1);
    }
  if (r < len)
    return r + SSL_write_all (ssl, data + len, len - r);
  else
    return r;
}



int main (int argc, char **argv)
{
  SSL *ssl = NULL;
  BIO *sbio = NULL;
  int sock = 0;
  int r = 0;
  SSL_METHOD *meth = NULL;
  SSL_CTX *ctx = NULL;
  int s_server_session_id_context = 1;
  char *host = "www.google.com";
  int port = 443;
  char *data_to_send =
    "GET / HTTP/1.1\r\n"
    "Host: www.google.com\r\n"
    "Accept: */*\r\n"
    "\r\n";
  int amount_read = 0;
  int chunks_read = 0;
  int do_shutdown_after_write = 1;

  SSL_library_init ();
  check_for_errors ("SSL_library_init");
  SSL_load_error_strings ();
  check_for_errors ("SSL_load_error_strings");

  meth = SSLv23_method ();
  check_for_errors ("SSLv23_method");
  /* http://www.openssl.org/docs/ssl/SSL_CTX_new.html */
  ctx = SSL_CTX_new (meth);
  check_for_errors ("SSL_CTX_new");

  SSL_CTX_set_session_id_context (ctx,
                                  (void*)&s_server_session_id_context,
                                  sizeof (s_server_session_id_context));
  check_for_errors ("SSL_CTX_set_session_id_context");

  sock=tcp_connect (host, port);

  ssl = SSL_new (ctx);
  check_for_errors ("SSL_new");
  sbio = BIO_new_socket (sock, BIO_NOCLOSE);

  check_for_errors ("BIO_new_socket");
  SSL_set_bio (ssl, sbio, sbio);
  check_for_errors ("SSL_set_bio");

  /*
  BIO_set_nbio (sbio, 0);
  check_for_errors ("BIO_set_nbio");
  */

  r = SSL_connect (ssl);
  check_for_errors ("SSL_connect");
  if (r <= 0)
    {
      printf ("SSL connect error\n");
      exit (1);
    }

  r = SSL_write_all (ssl, data_to_send, strlen (data_to_send));
  printf ("after write, r=%d\n", r);

  printf ("before first shutdown, SSL_get_shutdown=%s, SSL_state=%s\n",
          sd_state_to_string (SSL_get_shutdown (ssl)),
          SSL_state_string (ssl));

  if (do_shutdown_after_write)
    {
      /* http://www.openssl.org/docs/ssl/SSL_shutdown.html */
      r = SSL_shutdown (ssl);
      check_for_errors ("SSL_shutdown 0");
      printf ("SSL_shutdown returned %d\n", r);
    }

  printf ("after first shutdown, SSL_get_shutdown=%s, SSL_state=%s\n",
          sd_state_to_string (SSL_get_shutdown (ssl)),
          SSL_state_string (ssl));

  while (1)
    {
      char buffer[ 2048 ];
      r = SSL_read (ssl, buffer, sizeof (buffer) - 1);
      check_for_errors ("SSL_read");
      /* printf ("SSL_read returned %d\n", r); */
      if (r < 0)
        {
          printf ("error from SSL_read\n");
          exit (1);
        }
      if (r == 0)
        break;
      amount_read += r;
      chunks_read += 1;
      /*
        buffer[ r ] = '\0';
        printf ("%s", buffer);
      */
    }
  printf ("%d characters in %d chunks read from server.\n",
          amount_read, chunks_read);

  r = SSL_shutdown (ssl);
  printf ("SSL_shutdown returned %d\n", r);
  check_for_errors ("SSL_shutdown 1");
  printf ("after second shutdown, SSL_get_shutdown=%s, SSL_state=%s\n",
          sd_state_to_string (SSL_get_shutdown (ssl)),
          SSL_state_string (ssl));

  SSL_CTX_free (ctx);

  exit (0);
}
